//
//  BufferedReader.m
//  BufferedReader implementation
//
//  Created by Rene Hexel on 10/03/06.
//  Copyright 2006 Rene Hexel. All rights reserved.
//

#include <fcntl.h>
#import "BufferedReader.h"


@implementation BufferedReader

/*
 * convenience method that returns an autoreleased BufferedReader instance
 * with standard input as the source
 */
+ (BufferedReader *) bufferedReaderWithStandardInput
{
	return [[[BufferedReader alloc] initWithStandardInput] autorelease];
}


/*
 * default initializer
 */
- init
{
	if (!(self = [super init])) return nil;
	
	newLine = [[NSCharacterSet characterSetWithCharactersInString: @"\n\r"]
			retain];
	
	return self;
}


/*
 * initialise to read from a given file
 */
- initWithFile: (NSString *) fileName
{
	if (!(self = [self init])) return nil;

	fd = open([fileName cString], O_RDONLY, 0);
	if (fd < 0)
	{
		[self release];
		return nil;
	}

	file = [[NSFileHandle alloc] initWithFileDescriptor: fd
					     closeOnDealloc: YES];
	if (!file)
	{
		close(fd);
		[self release];
		return nil;
	}
	
	return self;
}


/*
 * initialise to read from standard input
 */
- (BufferedReader *) initWithStandardInput
{
	if (!(self = [self init])) return nil;
	
	file = [[NSFileHandle fileHandleWithStandardInput] retain];
	
	if (!file)
	{
		[self release];
		return nil;
	}

	return self;
}


/*
 * release referenced objects when deallocating
 */
- (void) dealloc
{
	[file release];
	[content release];
	[scanner release];
	[newLine release];

	[super dealloc];
}


/*
 * select whether to skip white space when scanning
 */
- setSkipWhiteSpace: (BOOL) skip
{
	skipWhiteSpace = skip;

	return self;
}


/*
 * read a single line from the input handle
 * this returns the line without the end of line character or nil if the end
 * of the input file has been reached
 */
- (NSString *) readLine
{
	NSString *line;

	/*
	 * if we haven't read anything yet, content is nil, so read
	 * available data from the input file and create a new scanner
	 */
	if (content == nil)
	{
		content = [[NSString alloc] initWithData: [file availableData]
						encoding: NSUTF8StringEncoding];

		if (content == nil) return nil;

		[scanner release];
		scanner = [[NSScanner scannerWithString: content] retain];

		// skip new line characters only?
		if (!skipWhiteSpace)
			[scanner setCharactersToBeSkipped: newLine];
	}

	/*
	 * if -isAtEnd is true, we either have read an empty line or exhausted
	 * the input.  In the latter case, we try -readLine again to see if new
	 * input data are available
	 */
	if ([scanner isAtEnd])
	{
		line = nil;

		if ([content isEqual: @"\n"] || [content isEqual: @"\r\n"] ||
		    [content isEqual: @"\r"] || [content isEqual: @"\n\r"])
			line = @"";
		else if ([content length] != 0)
		{
			[content release];
			content = nil;
			return [self readLine];
		}

		[content release];
		content = nil;
		return line;
	}
	else	/* scanner is not at end, so scan the line */
	if ([scanner scanUpToCharactersFromSet: newLine intoString: &line])
		return line;

	return nil;
}

@end
